@@ -61,7 +61,8 @@ replacements with certainty.  Anyway, iteration doesn't seem to be needed
 on - code for the moment.  Although it may be confusing that there can be
 iterated addition of code before context code where the context code is
 immediately followed by removed code. *)
-and adjacency = ALLMINUS | ADJ of int
+and adj = { counter : int; mutable ender : bool; }
+and adjacency = ALLMINUS | ADJ of adj
 and mcodekind =
     MINUS       of pos * int list * adjacency * anything replacement
   | CONTEXT     of pos * anything befaft
@@ -104,6 +105,7 @@ and metavar =
   | MetaParamListDecl of arity * meta_name (*name*) * list_len (*len*)
   | MetaBinaryOperatorDecl of arity * meta_name
   | MetaAssignmentOperatorDecl of arity * meta_name
+  | MetaPragmaInfoDecl of arity * meta_name
   | MetaConstDecl of
       arity * meta_name (* name *) * fullType list option
   | MetaErrDecl of arity * meta_name (* name *)
@@ -191,15 +193,25 @@ and base_expression =
   | Unary          of expression * unaryOp mcode
   | Binary         of expression * binaryOp * expression
   | Nested         of expression * binaryOp * expression
-  | ArrayAccess    of expression * string mcode (* [ *) * expression *
+  | ArrayAccess    of expression * string mcode (* [ *) * expression dots *
 	              string mcode (* ] *)
   | RecordAccess   of expression * string mcode (* . *) * ident
   | RecordPtAccess of expression * string mcode (* -> *) * ident
-  | Cast           of string mcode (* ( *) * fullType * attr list *
+  | QualifiedAccess of fullType option * string mcode (* :: *) * ident
+  | Cast           of string mcode (* ( *) * fullType *
                       string mcode (* ) *) * expression
   | SizeOfExpr     of string mcode (* sizeof *) * expression
   | SizeOfType     of string mcode (* sizeof *) * string mcode (* ( *) *
                       fullType * string mcode (* ) *)
+  | Delete         of string mcode (* delete *) * expression
+  | DeleteArr      of string mcode (* delete *) * string mcode (* [ *) * string mcode (* ] *) *
+	              expression
+  | New            of string mcode (* new *) * arguments option *
+	           string mcode (* ( *) option * fullType *
+	           string mcode (* ) *) option * arguments option
+  | TemplateInst   of expression (* name *) * string mcode (* < *) *
+        expression dots * string mcode (* > *)
+  | TupleExpr      of initialiser
   | TypeExp        of fullType (*type name used as an expression, only in
 				  arg or #define*)
 
@@ -237,6 +249,7 @@ and base_expression =
 and 'expression generic_constraints =
     CstrFalse
   | CstrTrue
+  | CstrExit (* only generated by asttoctl2 *)
   | CstrAnd of 'expression generic_constraints list
   | CstrOr of 'expression generic_constraints list
   | CstrNot of 'expression generic_constraints
@@ -276,6 +289,8 @@ and form = ANY | ID | LocalID | GlobalID | CONST (* form for MetaExp *)
 
 and expression = base_expression wrap
 
+and arguments = string mcode (* ( *) * expression dots * string mcode (* ) *)
+
 and listlen =
     MetaListLen of meta_name mcode * constraints * keep_binding * inherited
   | CstListLen of int
@@ -297,7 +312,8 @@ and base_string_format =
 
 and string_format = base_string_format wrap
 
-and unaryOp = GetRef | GetRefLabel | DeRef | UnPlus |  UnMinus | Tilde | Not
+and unaryOp =
+      GetRef | GetRefLabel | DeRef | UnPlus |  UnMinus | Tilde of string | Not of string
 and base_assignOp =
     SimpleAssign of simpleAssignOp mcode
   | OpAssign of arithOp mcode
@@ -314,9 +330,11 @@ and base_binaryOp =
       meta_name mcode * constraints * keep_binding * inherited
 and binaryOp = base_binaryOp wrap
 and arithOp =
-    Plus | Minus | Mul | Div | Mod | DecLeft | DecRight | And | Or | Xor
+    Plus | Minus | Mul | Div | Mod | DecLeft | DecRight
+  | And of string | Or of string | Xor of string
   | Min | Max
-and  logicalOp = Inf | Sup | InfEq | SupEq | Eq | NotEq | AndLog | OrLog
+and  logicalOp = Inf | Sup | InfEq | SupEq | Eq | NotEq of string
+  | AndLog of string | OrLog of string
 
 and constant =
     String of string * isWchar
@@ -330,8 +348,7 @@ and isWchar = IsWchar | IsUchar | Isuchar | Isu8char | IsChar
 (* Types *)
 
 and base_fullType =
-    Type            of bool (* true if all minus *) *
-	               const_vol mcode list * typeC
+    Type            of bool (* true if all minus *) * cvattr list * typeC * cvattr list
   | AsType          of fullType * fullType (* as type, always metavar *)
   | DisjType        of fullType list
   | ConjType        of fullType list
@@ -350,8 +367,8 @@ and base_typeC =
 	               expression *
 	               string mcode option (* , *) * expression option *
 	               string mcode (* ) *) (* IBM C only *)
-  | EnumName        of string mcode (*enum*) * ident option (* name *)
-  | EnumDef  of fullType (* either EnumName or metavar *) *
+  | EnumName        of string mcode (*enum*) * structUnion mcode option(*struct/class/None*) * ident option (* name *)
+  | EnumDef  of fullType (* either EnumName or metavar *) * enum_base option *
 	string mcode (* { *) * enum_decl dots * string mcode (* } *)
   | StructUnionName of structUnion mcode * ident option (* name *)
   | StructUnionDef  of fullType (* either StructUnionName or metavar *) *
@@ -361,7 +378,10 @@ and base_typeC =
   | TypeOfType      of string mcode (* sizeof *) * string mcode (* ( *) *
                        fullType * string mcode (* ) *)
   | TypeName        of string mcode (* pad: should be 'of ident' ? *)
+  | QualifiedType   of fullType option * string mcode (* :: *) * ident
   | AutoType        of string mcode (* auto *) (* c++ >= 11 *)
+  | TemplateType    of fullType (* name *) * string mcode (* < *) *
+        expression dots * string mcode (* > *)
 
   | MetaType        of meta_name mcode * constraints * keep_binding *
 	inherited
@@ -376,7 +396,7 @@ and baseType = VoidType | CharType | ShortType | ShortIntType | IntType
 | SizeType | SSizeType | PtrDiffType
 | BoolType | Unknown
 
-and structUnion = Struct | Union
+and structUnion = Struct | Union | Class
 
 and sign = Signed | Unsigned
 
@@ -388,23 +408,23 @@ and const_vol = Const | Volatile
    split out into multiple declarations of a single variable each. *)
 
 and base_declaration =
-  | Init of storage mcode option * fullType * attr list * ident *
-        attr list * string mcode (*=*) * initialiser * string mcode (*;*)
-  | UnInit of storage mcode option * fullType * attr list * ident *
+  | Init of alignas option * storage mcode option * fullType * ident *
+        attr list * string mcode (*=*) * initialiser * string mcode option (*;*)
+  | UnInit of alignas option * storage mcode option * fullType * ident *
 	attr list * string mcode (* ; *)
   | FunProto of
-	fninfo list * attr list * ident (* name *) *
+	fninfo list * ident (* name *) *
 	string mcode (* ( *) * parameter_list *
 	(string mcode (* , *) * string mcode (* ...... *) ) option *
 	string mcode (* ) *) * string mcode (* ; *)
-  | TyDecl of fullType * attr list * string mcode (* ; *)
-  | MacroDecl of storage mcode option *
+  | TyDecl of fullType * string mcode (* ; *)
+  | MacroDecl of storage mcode option * attr list *
 	ident (* name *) * string mcode (* ( *) *
         expression dots * string mcode (* ) *) *
         attr list * string mcode (* ; *)
-  | MacroDeclInit of storage mcode option *
+  | MacroDeclInit of storage mcode option * attr list *
 	ident (* name *) * string mcode (* ( *) *
-        expression dots * string mcode (* ) *) * string mcode (*=*) *
+        expression dots * string mcode (* ) *) * attr list * string mcode (*=*) *
         initialiser * string mcode (* ; *)
   | Typedef of string mcode (*typedef*) * fullType *
                typeC (* either TypeName or metavar *) * string mcode (*;*)
@@ -417,6 +437,9 @@ and base_declaration =
 
 and declaration = base_declaration wrap
 
+and alignas = Align of string mcode (* alignas *) * string mcode (* ( *)
+					* expression * string mcode (* ) *)
+
 and base_annotated_decl =
     DElem of mcodekind (* before the decl *) * bool (* true if all minus *) *
       declaration
@@ -427,10 +450,16 @@ and annotated_decl = base_annotated_decl wrap
 (* Field declaration *)
 
 and base_field =
-    Field of fullType * ident option * bitfield option * string mcode (* ; *)
+    Field of fullType * ident option * bitfield option * attr list * string mcode (* ; *)
+  | MacroDeclField of
+	ident (* name *) * string mcode (* ( *) *
+        expression dots * string mcode (* ) *) *
+        attr list * string mcode (* ; *)
+  | CppField  of directive
   | MetaField of meta_name mcode * constraints * keep_binding * inherited
   | MetaFieldList of meta_name mcode * listlen * constraints * keep_binding *
 	inherited
+  | AccSpec of string mcode * string mcode   (* C++ *)
 
 and bitfield = string mcode (* : *) * expression
 
@@ -451,6 +480,7 @@ and base_enum_decl =
   | EnumComma of string mcode (* , *)
   | EnumDots of string mcode (* ... *) * enum_decl option (* whencode *)
 
+and enum_base = string mcode (* : *) * fullType
 and enum_decl = base_enum_decl wrap
 
 (* --------------------------------------------------------------------- *)
@@ -489,8 +519,7 @@ and initialiser = base_initialiser wrap
 (* Parameter *)
 
 and base_parameterTypeDef =
-    VoidParam     of fullType * attr list
-  | Param         of fullType * attr list * ident option * attr list
+    Param         of fullType * ident option * attr list
 
   | MetaParam     of meta_name mcode * constraints * keep_binding * inherited
   | MetaParamList of meta_name mcode * listlen * constraints * keep_binding *
@@ -513,6 +542,7 @@ and parameter_list = parameterTypeDef dots
 
 and base_define_param =
     DParam        of ident
+  | DParamEll     of ident * string mcode (* ... *)
   | MetaDParamList of meta_name mcode * listlen * constraints * keep_binding *
 	inherited
   | DPComma       of string mcode
@@ -563,15 +593,13 @@ and base_rule_elem =
 	             expression * string mcode (* ) *)
   | Else          of string mcode (* else *)
   | WhileHeader   of string mcode (* while *) * string mcode (* ( *) *
-	             expression * string mcode (* ) *)
+	             whileinfo * string mcode (* ) *)
   | DoHeader      of string mcode (* do *)
   | WhileTail     of string mcode (* while *) * string mcode (* ( *) *
 	             expression * string mcode (* ) *) *
                      string mcode (* ; *)
   | ForHeader     of string mcode (* for *) * string mcode (* ( *) *
-                     forinfo *
-	             expression option * string mcode (*;*) *
-                     expression option * string mcode (* ) *)
+                     forinfo * string mcode (* ) *)
   | IteratorHeader of ident (* name *) * string mcode (* ( *) *
 	             expression dots * string mcode (* ) *)
   | SwitchHeader  of string mcode (* switch *) * string mcode (* ( *) *
@@ -598,38 +626,65 @@ and base_rule_elem =
   | Ty            of fullType (* only at SP top level, matches a subterm *)
   | TopId         of ident (* only at top level *)
   | TopInit       of initialiser (* only at top level *)
-  | Include       of string mcode (*#include*) * inc_file mcode (*file *)
-  | MetaInclude   of string mcode (* #include *) * expression (* file *)
+  | CppTop        of directive
   | Undef         of string mcode (* #define *) * ident (* name *)
   | DefineHeader  of string mcode (* #define *) * ident (* name *) *
 	             define_parameters (*params*)
-  | Pragma        of string mcode (* #pragma *) * ident * pragmainfo
+  | TemplateDefinitionHeader
+                  of string mcode (* template *) * string mcode (* < *) *
+                      template_parameter_list * string mcode (* > *)
   | Case          of string mcode (* case *) * expression * string mcode (*:*)
   | Default       of string mcode (* default *) * string mcode (*:*)
   | AsRe          of rule_elem * rule_elem (* always { and MetaStmtList *)
   | DisjRuleElem  of rule_elem list
 
+and base_templateParameterTypeDef =
+    TypenameOrClassParam of string mcode (* typename|class *) * ident (* name *) * (string mcode (* = *) * fullType) option
+  | VarNameParam of fullType * ident (* name *) * (string mcode (* = *) * initialiser) option
+  | TPComma of string mcode
+  | TPDots of string mcode (* ... *)
+  (* Note: TemplateParam not supported yet. *)
+and templateParameterTypeDef = base_templateParameterTypeDef wrap
+and template_parameter_list = templateParameterTypeDef dots
+
 and base_pragmainfo =
     PragmaString of string mcode
   | PragmaDots of string mcode
+  | MetaPragmaInfo of meta_name mcode * constraints * keep_binding * inherited
 
 and pragmainfo = base_pragmainfo wrap
 
 and forinfo =
-    ForExp of expression option * string mcode (*;*)
-  | ForDecl of annotated_decl
+    ForExp of expression option * string mcode (*;*) *
+	expression option * string mcode (*;*) *
+        expression option
+  | ForDecl of annotated_decl * expression option * string mcode (*;*) *
+        expression option
+  | ForRange of annotated_decl * initialiser
+
+and whileinfo =
+    WhileExp  of expression
+  | WhileDecl of annotated_decl
 
 and fninfo =
     FStorage of storage mcode
   | FType of fullType
   | FInline of string mcode
-  | FAttr of attr
+
+and cvattr =
+    CV of const_vol mcode
+  | Attr of attr
 
 and base_attr =
     Attribute of attr_arg
   | GccAttribute of string mcode (* __attribute__ *) *
                     string mcode (* ( *) * string mcode (* ( *) *
-                    attr_arg * string mcode (* ) *) * string mcode (* ) *)
+                    expression dots * string mcode (* ) *) * string mcode (* ) *)
+  | CxxAttribute of string mcode (* [[ *) *
+                    expression dots * string mcode (* ] *) * string mcode (* ] *)
+  | CxxAttributeUsing of string mcode (* [[ *) * string mcode (* using *) *
+                    ident * string mcode (* : *) *
+                    expression dots * string mcode (* ] *) * string mcode (* ] *)
 
 and attr = base_attr wrap
 
@@ -667,6 +722,7 @@ and base_statement =
 	             dots_whencode list * dots_whencode list
   | FunDecl       of rule_elem (* header *) * rule_elem (* { *) *
      	             statement dots * rule_elem (* } *) * end_info (*exit*)
+  | TemplateDefinition of rule_elem * statement
   | Define        of rule_elem (* header *) * statement dots
   | AsStmt        of statement * statement (* as statement, always metavar *)
   | Dots          of string mcode (* ... *) *
@@ -674,6 +730,20 @@ and base_statement =
 	             dots_whencode list * dots_whencode list
   | OptStm        of statement
 
+and base_directive =
+    Include       of string mcode (*#include*) * inc_file mcode (*file *)
+  | MetaInclude   of string mcode (* #include *) * expression (* file *)
+  | Pragma        of string mcode (* #pragma *) * ident * pragmainfo
+  | UsingNamespace of string mcode (*using*) * string mcode (*namespace*) *
+      ident (*name*) * string mcode (*;*)
+  | UsingTypename of string mcode (*using*) * ident (*name*) *
+      string mcode (*=*) * string mcode option (*typename*) *
+      fullType (*full_type*) * string mcode (*;*)
+  | UsingMember of string mcode (*using*) * ident (*name*) *
+      string mcode (*;*)
+
+and directive = base_directive wrap
+
 and ('a,'b) whencode =
     WhenNot of 'a
   | WhenAlways of 'b
@@ -824,6 +894,7 @@ and anything =
   | Code                of top_level
   | ExprDotsTag         of expression dots
   | ParamDotsTag        of parameterTypeDef dots
+  | TemplateParamDotsTag of templateParameterTypeDef dots
   | StmtDotsTag         of statement dots
   | AnnDeclDotsTag      of annotated_decl dots
   | AnnFieldDotsTag     of annotated_field dots
@@ -831,6 +902,7 @@ and anything =
   | DefParDotsTag       of define_param dots
   | TypeCTag            of typeC
   | ParamTag            of parameterTypeDef
+  | TemplateParamTag    of templateParameterTypeDef
   | SgrepStartTag       of string
   | SgrepEndTag         of string
 
@@ -859,11 +931,9 @@ let unwrap_mcode (x,_,_,_)  = x
 let get_mcodekind (_,_,x,_) = x
 let get_line x             = x.node_line
 let get_mcode_line (_,l,_,_) = l.line
-let get_mcode_col (_,l,_,_)  = l.column
 let get_fvs x              = x.free_vars
 let set_fvs fvs x          = {x with free_vars = fvs}
 let get_mfvs x             = x.minus_free_vars
-let set_mfvs mfvs x        = {x with minus_free_vars = mfvs}
 let get_minus_nc_fvs x     = x.minus_nc_free_vars
 let get_fresh x            = x.fresh_vars
 let get_inherited x        = x.inherited
@@ -909,6 +979,7 @@ let get_meta_name = function
   | MetaParamListDecl(_ar,nm,nm1) -> nm
   | MetaBinaryOperatorDecl(_,name) -> name
   | MetaAssignmentOperatorDecl(_,name) -> name
+  | MetaPragmaInfoDecl(_,name) -> name
   | MetaConstDecl(_ar,nm,_ty) -> nm
   | MetaErrDecl(_ar,nm) -> nm
   | MetaExpDecl(_ar,nm,_ty,_bitfield) -> nm
@@ -971,6 +1042,7 @@ and tag2c = function
   | Code _ -> "Code"
   | ExprDotsTag _ -> "ExprDotsTag"
   | ParamDotsTag _ -> "ParamDotsTag"
+  | TemplateParamDotsTag _ -> "TemplateParamDotsTag"
   | StmtDotsTag _ -> "StmtDotsTag"
   | AnnDeclDotsTag _ -> "AnnDeclDotsTag"
   | AnnFieldDotsTag _ -> "AnnFieldDotsTag"
@@ -978,6 +1050,7 @@ and tag2c = function
   | DefParDotsTag _ -> "DefParDotsTag"
   | TypeCTag _ -> "TypeCTag"
   | ParamTag _ -> "ParamTag"
+  | TemplateParamTag _ -> "TemplateParamTag"
   | SgrepStartTag _ -> "SgrepStartTag"
   | SgrepEndTag _ -> "SgrepEndTag"
 
@@ -1026,10 +1099,16 @@ let make_meta_rule_elem s d c (fvs,fresh,inh) =
       (MetaRuleElem(((rule,s),no_info,d,[]),c,Unitary,false)))
   with free_vars = fvs; fresh_vars = fresh; inherited = inh}
 
-let make_meta_decl s d c (fvs,fresh,inh) =
+let make_meta_id s d c (fvs,fresh,inh) =
   let rule = "" in
   {(make_term
-      (MetaDecl(((rule,s),no_info,d,[]),c,Unitary,false))) with
+      (MetaId(((rule,s),no_info,d,[]),c,Unitary,false))) with
+    free_vars = fvs; fresh_vars = fresh; inherited = inh}
+
+let make_meta_type s d c (fvs,fresh,inh) =
+  let rule = "" in
+  {(make_term
+      (MetaType(((rule,s),no_info,d,[]),c,Unitary,false))) with
     free_vars = fvs; fresh_vars = fresh; inherited = inh}
 
 let make_mcode x = (x,no_info,CONTEXT(NoPos,NOTHING),[])
@@ -1050,21 +1129,21 @@ let string_of_arithOp = function
   | Mod -> "%"
   | DecLeft -> "<<"
   | DecRight -> ">>"
-  | And -> "&"
-  | Or -> "|"
-  | Xor -> "^"
+  | And s -> s
+  | Or s -> s
+  | Xor s -> s
   | Min -> "<?"
   | Max -> ">?"
 
 let string_of_logicalOp = function
   | Eq -> "=="
-  | NotEq -> "!="
+  | NotEq s -> s
   | InfEq -> "<="
   | SupEq -> ">="
   | Sup -> ">"
   | Inf -> "<"
-  | AndLog -> "&&"
-  | OrLog -> "||"
+  | AndLog s -> s
+  | OrLog s -> s
 
 let string_of_binaryOp op = match (unwrap op) with
   | Arith arithOp -> string_of_arithOp (unwrap_mcode arithOp)
@@ -1104,8 +1183,9 @@ let string_of_baseType = function
   | BoolType -> "bool"
   | Unknown -> "unknown"
 
-let string_of_const_vol cv = String.concat " "
-  (List.map (function Const -> "const" | Volatile -> "volatile") cv)
+let string_of_const_vol = function
+    Const -> "const"
+  | Volatile -> "volatile"
 
 let string_of_meta_name (_, name) = name
 
@@ -1128,6 +1208,7 @@ let string_of_expression e =
 let string_of_structUnion = function
     Struct -> "struct"
   | Union -> "union"
+  | Class -> "class"
 
 let rec string_of_typeC ty =
   match unwrap ty with
@@ -1147,23 +1228,41 @@ let rec string_of_typeC ty =
       let s0 = string_of_expression e0
       and s1 = Common.default "?" string_of_expression e1 in
       Printf.sprintf "decimal(%s,%s) " s0 s1
-  | EnumName (_, name) -> "enum " ^ (Common.default "?" string_of_ident name)
+  | EnumName (_, key, name) ->
+      "enum " ^
+      (Common.default ""
+	 (function x -> string_of_structUnion (unwrap_mcode x)) key) ^
+      (Common.default ""
+	 (function x -> " ") key) ^ (Common.default "" string_of_ident name)
   | StructUnionName (kind, name) ->
       Printf.sprintf "%s %s"
 	(string_of_structUnion (unwrap_mcode kind))
 	(Common.default "?" string_of_ident name)
-  | EnumDef (ty', _, _, _)
+  | EnumDef (ty', _ , _, _, _)
   | StructUnionDef (ty', _, _, _) -> string_of_fullType ty'
   | TypeOfExpr(_,_,e,_) -> "typeof("^string_of_expression e^")"
   | TypeOfType(_,_,t,_) -> "typeof("^string_of_fullType t^")"
   | TypeName (name) -> unwrap_mcode name ^ " "
+  | TemplateType (name, _, args, _) ->
+      let args =
+	String.concat ", "
+	  (List.map string_of_expression (unwrap args)) in
+      Printf.sprintf "%s<%s>" (string_of_fullType name) args
+  | QualifiedType(Some ty,_,name) -> string_of_fullType ty ^ "::" ^ string_of_ident name ^ " "
+  | QualifiedType(None,_,name) -> "::" ^ string_of_ident name ^ " "
   | AutoType _ -> "auto"
   | MetaType (m, _, _, _) -> string_of_meta_name (unwrap_mcode m) ^ " "
 and string_of_fullType ty =
   match unwrap ty with
-    Type (_, [], ty') -> string_of_typeC ty'
-  | Type (_, const_vol, ty') ->
-      string_of_const_vol (List.map unwrap_mcode const_vol) ^ " " ^ string_of_typeC ty'
+    Type (_, cvbefore, ty', cvafter) ->
+      let do_cvattr = function
+	  CV cv -> string_of_const_vol (unwrap_mcode cv)
+	| Attr attr -> "TODO: attrs" in
+      (String.concat " " (List.map do_cvattr cvbefore)) ^
+      (if cvbefore = [] then "" else " ") ^
+      (string_of_typeC ty') ^
+      (if cvbefore = [] then "" else " ") ^
+      (String.concat " " (List.map do_cvattr cvafter))
   | AsType (ty', _) -> string_of_fullType ty'
   | DisjType l -> String.concat "|" (List.map string_of_fullType l)
   | ConjType l -> String.concat "&" (List.map string_of_fullType l)
@@ -1171,7 +1270,7 @@ and string_of_fullType ty =
 
 let typeC_of_fullType_opt ty =
   match unwrap ty with
-    Type (_, [], ty') -> Some ty'
+    Type (_, [], ty', []) -> Some ty'
   | _ -> None
 
 let ident_of_expression_opt expression =
@@ -1186,7 +1285,7 @@ type 'a transformer = {
     baseType: (baseType -> string mcode list -> 'a) option;
     decimal: (string mcode -> string mcode -> expression ->
       string mcode option -> expression option -> string mcode -> 'a) option;
-    enumName: (string mcode -> ident option -> 'a) option;
+    enumName: (string mcode -> structUnion mcode option -> ident option -> 'a) option;
     structUnionName: (structUnion mcode -> ident option -> 'a) option;
     typeName: (string mcode -> 'a) option;
     metaType: (meta_name mcode -> constraints -> keep_binding ->
@@ -1205,7 +1304,7 @@ let empty_transformer = {
 let rec fullType_map tr ty =
   rewrap ty begin
     match unwrap ty with
-      Type (a, b, ty') -> Type (a, b, typeC_map tr ty')
+      Type (a, b, ty', c) -> Type (a, b, typeC_map tr ty', c)
     | AsType (ty0, ty1) ->
         AsType (fullType_map tr ty0, fullType_map tr ty1)
     | DisjType l -> DisjType (List.map (fullType_map tr) l)
@@ -1227,11 +1326,11 @@ and typeC_map tr ty =
       rewrap ty (FunctionType (fullType_map tr ty', s0, s1, s2))
   | Array (ty', s0, s1, s2) ->
       rewrap ty (Array (fullType_map tr ty', s0, s1, s2))
-  | EnumName (s0, ident) ->
+  | EnumName (s0, key, ident) ->
       begin
         match tr.enumName with
           None -> ty
-        | Some f -> rewrap ty (f s0 ident)
+        | Some f -> rewrap ty (f s0 key ident)
       end
   | StructUnionName (su, ident) ->
       begin
@@ -1248,6 +1347,9 @@ and typeC_map tr ty =
           None -> ty
         | Some f -> rewrap ty (f name)
       end
+  | QualifiedType(Some ty',coloncolon,ident) ->
+      rewrap ty (QualifiedType (Some(fullType_map tr ty'), coloncolon, ident))
+  | QualifiedType(None,coloncolon,ident) -> ty
   | AutoType _ -> ty
   | MetaType (name, cstr, keep, inherited) ->
       begin
@@ -1264,14 +1366,16 @@ and typeC_map tr ty =
   | SignedT (_, None) -> ty
   | SignedT (sgn, Some ty') ->
       rewrap ty (SignedT (sgn, Some (typeC_map tr ty')))
-  | EnumDef (ty', s0, e, s1) ->
-      rewrap ty (EnumDef (fullType_map tr ty', s0, e, s1))
+  | EnumDef (ty', base, s0, e, s1) ->
+      rewrap ty (EnumDef (fullType_map tr ty', base, s0, e, s1))
   | StructUnionDef (ty', s0, a, s1) ->
       rewrap ty (StructUnionDef (fullType_map tr ty', s0, a, s1))
+  | TemplateType (nm, s0, s1, s2) ->
+      rewrap ty (TemplateType (nm, s0, s1, s2))
 
 let rec fullType_fold tr ty v =
   match unwrap ty with
-    Type (_, _, ty') -> typeC_fold tr ty' v
+    Type (_, _, ty', _) -> typeC_fold tr ty' v
   | AsType (ty0, ty1) ->
       let v' = fullType_fold tr ty0 v in
       fullType_fold tr ty1 v'
@@ -1287,26 +1391,27 @@ and typeC_fold tr ty v =
   | ParenType (_, ty', _)
   | FunctionType (ty', _, _, _)
   | Array (ty', _, _, _)
-  | EnumDef (ty', _, _, _)
+  | EnumDef (ty', _, _, _, _)
   | StructUnionDef (ty', _, _, _) -> fullType_fold tr ty' v
   | Decimal (s0, s1, e0, s2, e1, s3) ->
       Common.default v (fun f -> f s0 s1 e0 s2 e1 s3 v) tr.decimal
-  | EnumName (s0, ident) -> Common.default v (fun f -> f s0 ident v) tr.enumName
+  | EnumName (s0, key, ident) -> Common.default v (fun f -> f s0 key ident v) tr.enumName (* Not sure about this *)
   | StructUnionName (su, ident) ->
       Common.default v (fun f -> f su ident v) tr.structUnionName
   | TypeOfExpr(_,_,e,_) -> v
   | TypeOfType(_,_,t,_) -> fullType_fold tr t v
+  | TemplateType (_,_,_,_) -> v
   | TypeName name -> Common.default v (fun f -> f name v) tr.typeName
+  | QualifiedType(ty',_,ident) -> v
   | AutoType _ -> v
   | MetaType (name, cstr, keep, inherited) ->
       Common.default v (fun f -> f name cstr keep inherited v) tr.metaType
-
 let fullType_iter tr ty =
   fullType_fold {
     baseType = Common.map_option (fun f ty' s0 () -> f ty' s0) tr.baseType;
     decimal = Common.map_option
       (fun f s0 s1 e0 s2 e1 s3 () -> f s0 s1 e0 s2 e1 s3) tr.decimal;
-    enumName = Common.map_option (fun f s0 ident () -> f s0 ident) tr.enumName;
+    enumName = Common.map_option (fun f s0 key ident () -> f s0 key ident) tr.enumName;
     structUnionName = Common.map_option
       (fun f su ident () -> f su ident) tr.structUnionName;
     typeName = Common.map_option (fun f name () -> f name) tr.typeName;
@@ -1331,15 +1436,17 @@ let rec ident_fold_meta_names f ident v =
 let expression_fold_ident f e v = f (Common.just (ident_of_expression_opt e)) v
 
 let fullType_fold_meta_names f ty v =
-  let enumOrStructUnionName _ ident v =
+  let structUnionName _ ident v =
+    Common.default v (fun ident' -> ident_fold_meta_names f ident' v) ident in
+  let enumName _ _ ident v =
     Common.default v (fun ident' -> ident_fold_meta_names f ident' v) ident in
   fullType_fold { empty_transformer with
     decimal = Some (fun _ _ e1 _ e2 _ v ->
       let v' = expression_fold_ident (ident_fold_meta_names f) e1 v in
       Common.default v'
 	(fun e -> expression_fold_ident (ident_fold_meta_names f) e v) e2);
-    enumName = Some enumOrStructUnionName;
-    structUnionName = Some enumOrStructUnionName;
+    enumName = Some enumName;
+    structUnionName = Some structUnionName;
     metaType = Some (fun tyname _ _ _ v -> f (unwrap_mcode tyname) v)
   } ty v
 
@@ -1382,7 +1489,8 @@ let empty_cstr_transformer = {
 let rec cstr_fold_sign pos neg c accu =
   match c with
     CstrFalse
-  | CstrTrue -> accu
+  | CstrTrue
+  | CstrExit -> accu
   | CstrAnd list | CstrOr list ->
       List.fold_left (fun accu c' -> cstr_fold_sign pos neg c' accu) accu list
   | CstrNot c' -> cstr_fold_sign neg pos c' accu
@@ -1444,6 +1552,7 @@ let rec cstr_map transformer c =
   match c with
     CstrFalse -> CstrFalse
   | CstrTrue -> CstrTrue
+  | CstrExit -> CstrExit
   | CstrAnd list -> CstrAnd (List.map (cstr_map transformer) list)
   | CstrOr list -> CstrOr (List.map (cstr_map transformer) list)
   | CstrNot c' -> CstrNot (cstr_map transformer c')
