@@ -66,12 +66,6 @@ let unify_option f t1 t2 =
   | (None, None) -> true
   | _ -> false
 
-let unify_true_option f t1 t2 =
-  match (t1,t2) with
-    (Some t1, Some t2) -> f t1 t2
-  | (None, None) -> true
-  | _ -> true
-
 let bool_unify_option f t1 t2 =
   match (t1,t2) with
     (Some t1, Some t2) -> f t1 t2
@@ -214,24 +208,34 @@ and unify_expression e1 e2 =
       unify_binaryOp_mcode op1 op2 &&
       unify_expression l1 l2 &&
       unify_expression r1 r2
-  | (Ast.ArrayAccess(ar1,lb1,e1,rb1),Ast.ArrayAccess(ar2,lb2,e2,rb2)) ->
-      unify_expression ar1 ar2 && unify_expression e1 e2
+  | (Ast.ArrayAccess(f1,lb1,args1,rb1),Ast.ArrayAccess(f2,lb2,args2,rb2)) ->
+      unify_expression f1 f2 &&
+      unify_dots unify_expression edots args1 args2
   | (Ast.RecordAccess(e1,d1,fld1),Ast.RecordAccess(e2,d2,fld2)) ->
       unify_expression e1 e2 && unify_ident fld1 fld2
   | (Ast.RecordPtAccess(e1,pt1,fld1),Ast.RecordPtAccess(e2,pt2,fld2)) ->
       unify_expression e1 e2 && unify_ident fld1 fld2
-  | (Ast.Cast(lp1,ty1,attr1,rp1,e1),Ast.Cast(lp2,ty2,attr2,rp2,e2)) ->
-      if (List.length attr1 = List.length attr2) &&
-         List.for_all2 unify_attribute attr1 attr2
-      then unify_fullType ty1 ty2 && unify_expression e1 e2
-      else false
+  | (Ast.QualifiedAccess(ty1,coloncolon1,fld1),Ast.QualifiedAccess(ty2,coloncolon2,fld2)) ->
+      unify_option unify_fullType ty1 ty2 && unify_ident fld1 fld2
+  | (Ast.Cast(lp1,ty1,rp1,e1),Ast.Cast(lp2,ty2,rp2,e2)) ->
+      unify_fullType ty1 ty2 && unify_expression e1 e2
   | (Ast.SizeOfExpr(szf1,e1),Ast.SizeOfExpr(szf2,e2)) ->
       unify_expression e1 e2
   | (Ast.SizeOfType(szf1,lp1,ty1,rp1),Ast.SizeOfType(szf2,lp2,ty2,rp2)) ->
       unify_fullType ty1 ty2
+  | (Ast.Delete(dlt1,e1),Ast.Delete(dlt2,e2)) ->
+      unify_expression e1 e2
+  | (Ast.DeleteArr(dlt1,lb1,rb1,e1),Ast.DeleteArr(dlt2,lb2,rb2,e2))->
+      unify_expression e1 e2
+  | (Ast.New(new1,pp1_opt,lp1_opt,ty1,rp1_opt,args1_opt), Ast.New(new2,pp2_opt,lp2_opt,ty2,rp2_opt,args2_opt)) ->
+      unify_fullType ty1 ty2 &&
+      unify_option unify_args pp1_opt pp2_opt &&
+      unify_option unify_args args1_opt args2_opt
   | (Ast.TypeExp(ty1),Ast.TypeExp(ty2)) -> unify_fullType ty1 ty2
   | (Ast.Constructor(lp1,ty1,rp1,i1),Ast.Constructor(lp2,ty2,rp2,i2)) ->
       unify_fullType ty1 ty2 && unify_initialiser i1 i2
+  | Ast.TupleExpr(init1),Ast.TupleExpr(init2) ->
+      unify_initialiser init1 init2
   | (Ast.Paren(lp1,e1,rp1),Ast.Paren(lp2,e2,rp2)) ->
       unify_expression e1 e2
 
@@ -273,6 +277,10 @@ and unify_expression e1 e2 =
   | (_,Ast.OptExp(_)) -> failwith "unsupported expression"
   | _ -> false
 
+(* Used for the nonterminal arguments *)
+and unify_args (_,args1,_) (_,args2,_) =
+  unify_dots unify_expression edots args1 args2
+
 (* --------------------------------------------------------------------- *)
 (* Strings *)
 
@@ -299,9 +307,16 @@ and unify_string_format e1 e2 =
 
 and unify_fullType ft1 ft2 =
   match (Ast.unwrap ft1,Ast.unwrap ft2) with
-    (Ast.Type(_,cv1,ty1),Ast.Type(_,cv2,ty2)) ->
-      if List.length cv1 = List.length cv2 &&
-         List.for_all2 unify_mcode cv1 cv2
+    (Ast.Type(_,cv1before,ty1,cv1after),Ast.Type(_,cv2before,ty2,cv2after)) ->
+      let do_cvattr cvattr1 cvattr2 =
+	match (cvattr1,cvattr2) with
+	  (Ast.CV cv1, Ast.CV cv2) -> unify_mcode cv1 cv2
+	| (Ast.Attr attr1, Ast.Attr attr2) -> unify_attribute attr1 attr2
+	| _ -> false in
+      if List.length cv1before = List.length cv2before &&
+	 List.for_all2 do_cvattr cv1before cv2before &&
+	 List.length cv1after = List.length cv2after &&
+	 List.for_all2 do_cvattr cv1after cv2after
       then unify_typeC ty1 ty2
       else false
   | (Ast.AsType(ty1,asty1),_) ->
@@ -351,13 +366,14 @@ and unify_typeC t1 t2 =
      Ast.Decimal(dec2,lp2,len2,comma2,prec_opt2,rp2)) ->
       unify_expression len1 len2 &&
       unify_option unify_expression prec_opt1 prec_opt2
-  | (Ast.EnumName(s1,Some ts1),Ast.EnumName(s2,Some ts2)) ->
-      if unify_mcode s1 s2 then unify_ident ts1 ts2 else false
-  | (Ast.EnumName(s1,None),Ast.EnumName(s2,None)) ->
-      true
-  | (Ast.EnumDef(ty1,lb1,ids1,rb1),Ast.EnumDef(ty2,lb2,ids2,rb2)) ->
+  | (Ast.EnumName(s1,key1,Some ts1),Ast.EnumName(s2,key2,Some ts2)) ->
+      if (unify_mcode s1 s2) && (unify_option unify_mcode key1 key2) then unify_ident ts1 ts2 else false
+  | (Ast.EnumName(s1,key1,None),Ast.EnumName(s2,key2,None)) ->
+      unify_option unify_mcode key1 key2
+  | (Ast.EnumDef(ty1,base1,lb1,ids1,rb1),Ast.EnumDef(ty2,base2,lb2,ids2,rb2)) ->
       unify_fullType ty1 ty2 &&
-      unify_dots unify_enum_decl enumdots ids1 ids2
+      unify_option (fun (_, ty1) (_,ty2) -> unify_fullType ty1 ty2) base1 base2 &&
+      (unify_dots unify_enum_decl enumdots) ids1 ids2
   | (Ast.StructUnionName(s1,Some ts1),Ast.StructUnionName(s2,Some ts2)) ->
       if unify_mcode s1 s2 then unify_ident ts1 ts2 else false
   | (Ast.StructUnionName(s1,None),Ast.StructUnionName(s2,None)) ->
@@ -370,6 +386,10 @@ and unify_typeC t1 t2 =
       unify_expression e1 e2
   | (Ast.TypeOfType(szf1,lp1,ty1,rp1),Ast.TypeOfType(szf2,lp2,ty2,rp2)) ->
       unify_fullType ty1 ty2
+  | (Ast.QualifiedType(Some ty1,coloncolon1,fld1),Ast.QualifiedType(Some ty2,coloncolon2,fld2)) ->
+      unify_fullType ty1 ty2 && unify_ident fld1 fld2
+  | (Ast.QualifiedType(None,coloncolon,fld1),Ast.QualifiedType(None,coloncolon2,fld2)) ->
+      unify_ident fld1 fld2
   | (Ast.TypeName(t1),Ast.TypeName(t2)) -> unify_mcode t1 t2
   | (Ast.AutoType(auto1), Ast.AutoType(auto2)) -> unify_mcode auto1 auto2
 
@@ -382,31 +402,31 @@ and unify_typeC t1 t2 =
 (* Even if the Cocci program specifies a list of declarations, they are
    split out into multiple declarations of a single variable each. *)
 
+and unify_alignas (Ast.Align(al1,lp1,expr1,rp1)) (Ast.Align(al2,lp2,expr2,rp2)) =
+      unify_expression expr1 expr2
+
 and unify_declaration d1 d2 =
   match (Ast.unwrap d1,Ast.unwrap d2) with
     (Ast.MetaDecl(_,_,_,_),_) | (_,Ast.MetaDecl(_,_,_,_)) -> true
-  | (Ast.Init(stg1,ft1,midattr1,id1,endattr1,eq1,i1,s1),
-     Ast.Init(stg2,ft2,midattr2,id2,endattr2,eq2,i2,s2)) ->
+  | (Ast.Init(al1,stg1,ft1,id1,endattr1,eq1,i1,s1),
+     Ast.Init(al2,stg2,ft2,id2,endattr2,eq2,i2,s2)) ->
       if bool_unify_option unify_mcode stg1 stg2 &&
-         (List.length midattr1 = List.length midattr2) &&
-         List.for_all2 unify_attribute midattr1 midattr2 &&
          (List.length endattr1 = List.length endattr2) &&
          List.for_all2 unify_attribute endattr1 endattr2
       then
+	bool_unify_option unify_alignas al1 al2 &&
 	unify_fullType ft1 ft2 &&
 	unify_ident id1 id2 &&
 	unify_initialiser i1 i2
       else false
-  | (Ast.UnInit(stg1,ft1,midattr1,id1,endattr1,s1),Ast.UnInit(stg2,ft2,midattr2,id2,endattr2,s2)) ->
+  | (Ast.UnInit(al1,stg1,ft1,id1,endattr1,s1),Ast.UnInit(al2,stg2,ft2,id2,endattr2,s2)) ->
       if bool_unify_option unify_mcode stg1 stg2 &&
-         (List.length midattr1 = List.length midattr2) &&
-         List.for_all2 unify_attribute midattr1 midattr2 &&
          (List.length endattr1 = List.length endattr2) &&
          List.for_all2 unify_attribute endattr1 endattr2
-      then unify_fullType ft1 ft2 && unify_ident id1 id2
+      then bool_unify_option unify_alignas al1 al2 && unify_fullType ft1 ft2 && unify_ident id1 id2
       else false
-  | (Ast.FunProto(fi1,attr1,nm1,lp1,params1,va1,rp1,sem1),
-     Ast.FunProto(fi2,attr2,nm2,lp2,params2,va2,rp2,sem2)) ->
+  | (Ast.FunProto(fi1,nm1,lp1,params1,va1,rp1,sem1),
+     Ast.FunProto(fi2,nm2,lp2,params2,va2,rp2,sem2)) ->
        let l1 = match va1 with
          | None -> [lp1;rp1]
          | Some (c1,e1) -> [lp1;c1;e1;rp1] in
@@ -416,33 +436,34 @@ and unify_declaration d1 d2 =
        if List.for_all2 unify_mcode l1 l2
        then
 	  unify_fninfo fi1 fi2 &&
-          List.for_all2 unify_attribute attr1 attr2 &&
-          (List.length attr1 = List.length attr2) &&
 	  unify_ident nm1 nm2 &&
 	  unify_dots unify_parameterTypeDef pdots params1 params2
        else false
-  | (Ast.MacroDecl(s1,n1,lp1,args1,rp1,attr1,sem1),
-     Ast.MacroDecl(s2,n2,lp2,args2,rp2,attr2,sem2)) ->
+  | (Ast.MacroDecl(s1,preattr1,n1,lp1,args1,rp1,attr1,sem1),
+     Ast.MacroDecl(s2,preattr2,n2,lp2,args2,rp2,attr2,sem2)) ->
        if bool_unify_option unify_mcode s1 s2 &&
+         (List.length preattr1 = List.length preattr2) &&
+         List.for_all2 unify_attribute preattr1 preattr2 &&
          (List.length attr1 = List.length attr2) &&
          List.for_all2 unify_attribute attr1 attr2
        then
 	 unify_ident n1 n2 &&
 	 unify_dots unify_expression edots args1 args2
        else false
-  | (Ast.MacroDeclInit(s1,n1,lp1,args1,rp1,eq1,ini1,sem1),
-     Ast.MacroDeclInit(s2,n2,lp2,args2,rp2,eq2,ini2,sem2)) ->
-       if bool_unify_option unify_mcode s1 s2
+  | (Ast.MacroDeclInit(s1,preattr1,n1,lp1,args1,rp1,attr1,eq1,ini1,sem1),
+     Ast.MacroDeclInit(s2,preattr2,n2,lp2,args2,rp2,attr2,eq2,ini2,sem2)) ->
+       if bool_unify_option unify_mcode s1 s2 &&
+         (List.length preattr1 = List.length preattr2) &&
+         List.for_all2 unify_attribute preattr1 preattr2 &&
+         (List.length attr1 = List.length attr2) &&
+         List.for_all2 unify_attribute attr1 attr2
        then
 	 unify_ident n1 n2 &&
 	 unify_dots unify_expression edots args1 args2 &&
 	 unify_initialiser ini1 ini2
        else false
-  | (Ast.TyDecl(ft1,attr1,s1),Ast.TyDecl(ft2,attr2,s2)) ->
-      if (List.length attr1 = List.length attr2) &&
-         List.for_all2 unify_attribute attr1 attr2
-      then unify_fullType ft1 ft2
-      else false
+  | (Ast.TyDecl(ft1,s1),Ast.TyDecl(ft2,s2)) ->
+      unify_fullType ft1 ft2
   | (Ast.Typedef(stg1,ft1,id1,s1),Ast.Typedef(stg2,ft2,id2,s2)) ->
       unify_fullType ft1 ft2 && unify_typeC id1 id2
   | (Ast.DisjDecl(d1),_) ->
@@ -474,11 +495,22 @@ and unify_field d1 d2 =
     (Ast.MetaField(_,_,_,_),_) | (_,Ast.MetaField(_,_,_,_)) -> true
   | (Ast.MetaFieldList(_,_,_,_,_),_) | (_,Ast.MetaFieldList(_,_,_,_,_)) ->
       true
-  | (Ast.Field(ft1,id1,bf1,s1),Ast.Field(ft2,id2,bf2,s2)) ->
-      let unify_bitfield (c1, e1) (c2, e2) =
+  | (Ast.Field(ft1,id1,bf1,endattr1,s1),Ast.Field(ft2,id2,bf2,endattr2,s2)) ->
+      (List.length endattr1 = List.length endattr2) &&
+      List.for_all2 unify_attribute endattr1 endattr2 &&
+      (let unify_bitfield (c1, e1) (c2, e2) =
 	unify_mcode c1 c2 && unify_expression e1 e2 in
       unify_fullType ft1 ft2 && unify_option unify_ident id1 id2 &&
-      unify_option unify_bitfield bf1 bf2
+      unify_option unify_bitfield bf1 bf2)
+  | (Ast.MacroDeclField(n1,lp1,args1,rp1,attr1,sem1),
+     Ast.MacroDeclField(n2,lp2,args2,rp2,attr2,sem2)) ->
+       if (List.length attr1 = List.length attr2) &&
+         List.for_all2 unify_attribute attr1 attr2
+       then
+	 unify_ident n1 n2 &&
+	 unify_dots unify_expression edots args1 args2
+       else false
+  | _ -> false
 
 and unify_annotated_field d1 d2 =
   match (Ast.unwrap d1,Ast.unwrap d2) with
@@ -560,16 +592,9 @@ and unify_designator d1 d2 =
 
 and unify_parameterTypeDef p1 p2 =
   match (Ast.unwrap p1,Ast.unwrap p2) with
-    (Ast.VoidParam(ft1,attr1),Ast.VoidParam(ft2,attr2)) ->
+    (Ast.Param(ft1,i1,attr1),Ast.Param(ft2,i2,attr2)) ->
       if (List.length attr1 = List.length attr2) &&
          List.for_all2 unify_attribute attr1 attr2
-      then unify_fullType ft1 ft2
-      else false
-  | (Ast.Param(ft1,mattr1,i1,attr1),Ast.Param(ft2,mattr2,i2,attr2)) ->
-      if (List.length mattr1 = List.length mattr2) &&
-         List.for_all2 unify_attribute mattr1 mattr2 &&
-         (List.length attr1 = List.length attr2) &&
-         List.for_all2 unify_attribute attr1 attr2
       then
         unify_fullType ft1 ft2 &&
         unify_option unify_ident i1 i2
@@ -606,6 +631,8 @@ and unify_define_param p1 p2 =
   match (Ast.unwrap p1,Ast.unwrap p2) with
     (Ast.DParam(i1),Ast.DParam(i2)) ->
 	(unify_ident i1 i2)
+  | (Ast.DParamEll(i1,dots1),Ast.DParamEll(i2,dots2)) ->
+	(unify_ident i1 i2)
   | (Ast.MetaDParamList(_,_,_,_,_),_)
   | (_,Ast.MetaDParamList(_,_,_,_,_)) -> true
   | (Ast.DPComma(_),Ast.DPComma(_)) -> true
@@ -620,6 +647,12 @@ and unify_define_param p1 p2 =
 (* --------------------------------------------------------------------- *)
 (* Top-level code *)
 
+and unify_whileinfo c1 c2 =
+  match (c1,c2) with
+      (Ast.WhileExp(e1),Ast.WhileExp(e2)) -> unify_expression e1 e2
+    | (Ast.WhileDecl(d1),Ast.WhileDecl(d2)) -> unify_annotated_decl d1 d2
+    | _ -> false
+
 and unify_rule_elem re1 re2 =
   match (Ast.unwrap re1,Ast.unwrap re2) with
     (Ast.FunHeader(_,_,fi1,nm1,lp1,params1,va1,rp1,attrs1),
@@ -639,22 +672,24 @@ and unify_rule_elem re1 re2 =
   | (Ast.IfHeader(if1,lp1,e1,rp1),Ast.IfHeader(if2,lp2,e2,rp2)) ->
       unify_expression e1 e2
   | (Ast.Else(e1),Ast.Else(e2)) -> true
-  | (Ast.WhileHeader(wh1,lp1,e1,rp1),Ast.WhileHeader(wh2,lp2,e2,rp2)) ->
-      unify_expression e1 e2
+  | (Ast.WhileHeader(wh1,lp1,c1,rp1),Ast.WhileHeader(wh2,lp2,c2,rp2)) ->
+      unify_whileinfo c1 c2
   | (Ast.DoHeader(d1),Ast.DoHeader(d2)) -> true
   | (Ast.WhileTail(wh1,lp1,e1,rp1,s1),Ast.WhileTail(wh2,lp2,e2,rp2,s2)) ->
       unify_expression e1 e2
-  | (Ast.ForHeader(fr1,lp1,first1,e21,s21,e31,rp1),
-     Ast.ForHeader(fr2,lp2,first2,e22,s22,e32,rp2)) ->
-       let first =
-	 match (first1,first2) with
-	   (Ast.ForExp(e11,s11),Ast.ForExp(e12,s1)) ->
-	     unify_option unify_expression e11 e12
-	 | (Ast.ForDecl d1,Ast.ForDecl d2) -> unify_annotated_decl d1 d2
-	 | _ -> false in
-       first &&
-       unify_option unify_expression e21 e22 &&
-       unify_option unify_expression e31 e32
+  | (Ast.ForHeader(fr1,lp1,first1,rp1),Ast.ForHeader(fr2,lp2,first2,rp2)) ->
+       (match (first1,first2) with
+	 (Ast.ForExp(e11,s11,e21,s21,e31),Ast.ForExp(e12,s12,e22,s22,e32)) ->
+	   unify_option unify_expression e11 e12 &&
+	   unify_option unify_expression e21 e22 &&
+	   unify_option unify_expression e31 e32
+       | (Ast.ForDecl(d1,e21,s21,e31),Ast.ForDecl(d2,e22,s22,e32)) ->
+	   unify_annotated_decl d1 d2 &&
+	   unify_option unify_expression e21 e22 &&
+	   unify_option unify_expression e31 e32
+       | (Ast.ForRange(d1,i1),Ast.ForRange(d2,i2)) ->
+	   unify_annotated_decl d1 d2 && unify_initialiser i1 i2
+       | _ -> false)
   | (Ast.IteratorHeader(nm1,lp1,args1,rp1),
      Ast.IteratorHeader(nm2,lp2,args2,rp2)) ->
        unify_ident nm1 nm2 &&
@@ -663,8 +698,6 @@ and unify_rule_elem re1 re2 =
   | (Ast.DefineHeader(_,n1,p1),Ast.DefineHeader(_,n2,p2)) ->
        unify_ident n1 n2 &&
        unify_define_parameters p1 p2
-  | (Ast.Pragma(_,i1,n1),Ast.Pragma(_,i2,n2)) ->
-      unify_ident i1 i2 && unify_pragmainfo n1 n2
   | (Ast.Break(r1,s1),Ast.Break(r2,s2)) -> true
   | (Ast.Continue(r1,s1),Ast.Continue(r2,s2)) -> true
   | (Ast.Label(l1,dd1),Ast.Label(l2,dd2)) -> unify_ident l1 l2
@@ -705,11 +738,6 @@ and unify_rule_elem re1 re2 =
   | (_,Ast.Ty(t2)) -> subtype (unify_fullType t2) re1
   | _ -> false
 
-and unify_pragmainfo pi1 pi2 =
-  match (Ast.unwrap pi1,Ast.unwrap pi2) with
-      (Ast.PragmaString(s1),Ast.PragmaString(s2)) -> unify_mcode s1 s2
-    | (Ast.PragmaDots(_),_) | (_,Ast.PragmaDots(_)) -> true
-
 and unify_fninfo patterninfo cinfo =
   let patterninfo = List.sort compare patterninfo in
   let cinfo = List.sort compare cinfo in
@@ -720,8 +748,6 @@ and unify_fninfo patterninfo cinfo =
 	unify_fullType tya tyb && loop (resta,restb)
     | (Ast.FInline(ia)::resta,Ast.FInline(ib)::restb) ->
 	if unify_mcode ia ib then loop (resta,restb) else false
-    | (Ast.FAttr(ia)::resta,Ast.FAttr(ib)::restb) ->
-	if unify_attribute ia ib then loop (resta,restb) else false
     | (x::resta,((y::_) as restb)) ->
 	(match compare x y with
 	  -1 -> false
@@ -732,9 +758,14 @@ and unify_fninfo patterninfo cinfo =
 
 and unify_attribute attr1 attr2 =
   match (Ast.unwrap attr1,Ast.unwrap attr2) with
-    (Ast.Attribute(arg1),Ast.Attribute(arg2))
-  | (Ast.GccAttribute(_,_,_,arg1,_,_),Ast.GccAttribute(_,_,_,arg2,_,_)) ->
+    (Ast.Attribute(arg1),Ast.Attribute(arg2)) ->
       unify_attr_arg arg1 arg2
+  | (Ast.GccAttribute(_,_,_,args1,_,_),Ast.GccAttribute(_,_,_,args2,_,_)) ->
+      unify_dots unify_expression edots args1 args2
+  | (Ast.CxxAttribute(_,args1,_,_),Ast.CxxAttribute(_,args2,_,_)) ->
+      unify_dots unify_expression edots args1 args2
+  | (Ast.CxxAttributeUsing(_,_,atnm1,_,args1,_,_),Ast.CxxAttributeUsing(_,_,atnm2,_,args2,_,_)) ->
+      unify_ident atnm1 atnm2 && unify_dots unify_expression edots args1 args2
   | _ -> false
 
 and unify_attr_arg arg1 arg2 =
@@ -763,14 +794,9 @@ and subexp f =
   let mcode r e = option_default in
   let expr r k e = f e && k e in
   let donothing r k e = k e in
-  let recursor = V.combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing expr
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing donothing in
+  let recursor =
+    V.combiner bind option_default {V.cmcode=mcode} {V.cdonothing=donothing}
+      ~expr:expr donothing in
   recursor.V.combiner_rule_elem
 
 and subtype f =
@@ -779,14 +805,9 @@ and subtype f =
   let mcode r e = option_default in
   let fullType r k e = f e && k e in
   let donothing r k e = k e in
-  let recursor = V.combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing fullType
-      donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing in
+  let recursor =
+    V.combiner bind option_default {V.cmcode=mcode} {V.cdonothing=donothing}
+      ~ft:fullType donothing in
   recursor.V.combiner_rule_elem
 
 let rec unify_statement s1 s2 =
